from python_search.exceptions import CommandDoNotMatchException
from python_search.interpreter.base import BaseInterpreter


class PythonInterpreter(BaseInterpreter):
    def __init__(self, cmd, context=None):

        self.context = context

        if type(cmd) is dict and "callable" in cmd:
            self.cmd = cmd
            return

        raise CommandDoNotMatchException(f"Not Valid Python command {cmd}")

    def interpret_default(self):
        return self.cmd["callable"]

    def serialize(self):
        import dill

        return dill.source.getsource(self.cmd["callable"])



import os

from python_search.apps.terminal import Terminal
from python_search.config import PythonSearchConfiguration
from python_search.environment import is_mac
from python_search.observability.logger import logging


class FzfInTerminal:
    """
    Renders the search ui using fzf + termite terminal
    """

    FONT_SIZE = 14
    PREVIEW_PERCENTAGE_SIZE = 50
    HEIGHT = 240
    WIDTH = 950

    configuration: PythonSearchConfiguration

    @staticmethod
    def build_search_ui(configuration) -> "FzfInTerminal":
        """Assembles what is specific for the search ui exclusively"""

        return FzfInTerminal(configuration=configuration)

    def __init__(self, *, configuration: PythonSearchConfiguration):
        self.height = FzfInTerminal.HEIGHT
        self.width = FzfInTerminal.WIDTH

        self.preview_cmd = f"python_search _preview_entry {{}} "
        self.executable = "python_search"
        self.title = configuration.APPLICATION_TITLE
        self.configuration = configuration

    def run(self) -> None:
        self._launch_terminal(self.internal_cmd())

    def internal_cmd(self):
        FZF_LIGHT_THEME = "fg:#4d4d4c,bg:#ffffff,hl:#d7005f,info:#4271ae,prompt:#8959a8,pointer:#d7005f,marker:#4271ae,spinner:#4271ae,header:#4271ae,fg+:#4d4d4c,bg+:#e8e8e8,hl+:#d7005f"
        # for more fzf options see: https://www.mankier.com/1/fzf#
        cmd = f"""bash -c '{self._get_rankging_generate_cmd()} | \
        fzf \
        --tiebreak=length,begin,index \
        --cycle \
        --no-hscroll \
        --hscroll-off=0 \
        --preview "{self.preview_cmd}" \
        --preview-window=right,{FzfInTerminal.PREVIEW_PERCENTAGE_SIZE}%,wrap,border-left \
        --reverse -i --exact --no-sort \
        --border=none \
        --margin=0% \
        --padding=0% \
        {self._run_key("enter")} \
        {self._run_key("double-click")} \
        {self._edit_key('ctrl-e')} \
        {self._edit_key('right-click')} \
        --bind "alt-enter:execute-silent:(nohup {self.executable} run_key {{}} --query_used {{q}} & disown)" \
        --bind "alt-m:execute-silent:(nohup {self.executable} edit_main {{}} & disown)" \
        --bind "alt-m:+execute-silent:({self.executable} _utils hide_launcher)" \
        --bind "ctrl-l:clear-query" \
        --bind "ctrl-l:+first" \
        --bind "ctrl-k:execute-silent:(nohup {self.executable} copy_key_only {{}} & disown)" \
        --bind "ctrl-c:execute-silent:(nohup {self.executable} copy_entry_content {{}} & disown)" \
        --bind "ctrl-c:+execute-silent:({self.executable} _utils hide_launcher)" \
        --bind "ctrl-s:execute-silent:(nohup {self.executable} search_edit {{}} & disown)" \
        --bind "ctrl-s:+execute-silent:({self.executable} _utils hide_launcher)" \
        --bind "ctrl-r:reload:({self._get_rankging_generate_cmd(reload=True)})" \
        --bind "ctrl-t:execute-silent:(notify-send testjean)" \
        --bind "alt-g:execute-silent:( {self.executable} google_it {{q}} )" \
        --bind "alt-g:+execute-silent:({self.executable} _utils hide_launcher)" \
        --bind "ctrl-g:execute-silent:( {self.executable} google_it {{q}} )" \
        --bind "ctrl-g:+execute-silent:({self.executable} _utils hide_launcher)" \
        --bind "ctrl-f:first" \
        --bind "shift-up:first" \
        --bind "esc:abort" \
        --bind "ctrl-d:abort"  \
        --color={FZF_LIGHT_THEME}  ; exit 0
        '
        """

        logging.debug("Cmd: ", cmd)
        return cmd

    def _run_key(self, shortcut) -> str:
        return f"""--bind "{shortcut}:execute-silent:(LOG_FILE=/tmp/log_run_key_fzf nohup  log_command.sh {self.executable} run_key {{}} --query_used {{q}} & disown)" \
        --bind "{shortcut}:+execute-silent:({self.executable} _utils hide_launcher)" \
        --bind "{shortcut}:+reload:({self._get_rankging_generate_cmd()})" \
        --bind "{shortcut}:+clear-query" \
        --bind "{shortcut}:+abort"   """

    def _edit_key(self, shortcut) -> str:
        return f"""--bind "{shortcut}:execute-silent:(nohup {self.executable} edit_key {{}} & disown)" \
        --bind "{shortcut}:+execute-silent:({self.executable} _utils hide_launcher)"  """

    def _get_rankging_generate_cmd(self, reload=False):
        # in mac we need tensorflow to be installed via conda
        if self.configuration.supported_features.is_dynamic_ranking_supported():
            if reload:
                return f"curl -s localhost:8000/ranking/reload_and_generate"

            return f"curl -s localhost:8000/ranking/generate"

        return f"{self.executable} ranking generate"

    def _launch_terminal(self, internal_cmd: str) -> None:

        font = "FontAwesome"
        if is_mac():
            font = "Monaco"

        # --start-as=fullscreen \
        launch_cmd = f"""nice -19 kitty \
        --title="{self.title}"\
        -o initial_window_width={self.width}  \
        -o initial_window_height={self.height} \
        -o font_family="{font}" \
        -o font_size={FzfInTerminal.FONT_SIZE} \
        {Terminal.GLOBAL_TERMINAL_PARAMS} \
         {internal_cmd}
        """
        logging.info(f"Command performed:\n {internal_cmd}")
        result = os.system(launch_cmd)
        if result != 0:
            raise Exception("Search run fzf projection failed")


if __name__ == "__main__":
    import fire

    fire.Fire()



import streamlit as st

from python_search.ranking.next_item_predictor.training_dataset import \
    TrainingDataset


def load_training_page():
    st.write("## Training Dataset")

    df = load_dataset()
    pdf = df.toPandas()
    st.dataframe(pdf)

    st.write("### Pandas describe")
    st.write(pdf.describe())
    import matplotlib.pyplot as plt

    st.write("### Label histogram")
    fig, ax = plt.subplots()
    ax.hist(pdf["label"], bins=20)
    st.pyplot(fig)


def load_dataset():
    df = TrainingDataset().build(use_cache=True)
    return df



import os

from python_search.environment import is_mac


class Project:
    def new(self, new_project_location):
        """
        Initialize a new project to use Python search and make sure all remaining dependencies exist
        """

        if os.path.exists(new_project_location):
            raise Exception(f"{new_project_location} already exists")

        os.system(f"mkdir -p {new_project_location} 2>/dev/null")

        script_dir = os.path.dirname(os.path.realpath(__file__))
        copy_cmd = f"cp -r {script_dir}/entries_main.py {new_project_location}"
        os.system(copy_cmd)
        os.system(f"cd {new_project_location} && git init . 1>/dev/null ")

        self._install_kitty()
        self._install_fzf()
        self._set_current_project(new_project_location)

        print(
            f"""Project created successfully! 

Your main config script can be found at {new_project_location}/entries_main.py

You can now start using python search by issuing:
python_search search"""
        )

    def _set_current_project(self, project_directory):
        os.system("mkdir -p ~/.config/python_search/")
        os.system(
            f'echo "{project_directory}" >  ~/.config/python_search/current_project'
        )

    def _install_fzf(self):
        result = os.system("which fzf >/dev/null")
        if result == 0:
            print("Great, you have fzf alread installed")
            return

        print("Looks like kitty is not installed in your platform. ")
        if is_mac():
            print("Installing it for you...")
            os.system("brew install fzf")
        else:
            print(
                "Dont know how to install fzf for your platform, please do so manually"
            )

    def _install_kitty(self):
        result = os.system("which kitty >/dev/null")
        if result != 0:
            print(
                "Looks like kitty is not installed in your platform. Installing it for you..."
            )
            os.system(
                "curl -L https://sw.kovidgoyal.net/kitty/installer.sh | sh /dev/stdin"
            )



import logging
import os

from grimoire.file import write_file
from grimoire.shell import shell

from python_search.entries_group import EntriesGroup


class I3:
    """
    Setup a custom way of dealing with i3 shortcuts
    For this to work you have to rename your $HOME/.config/i3/config to $HOME/.config/i3/config_part 1
    And then let this script generate the file $HOME/.config/i3/config
    """

    def __init__(self, configuration: EntriesGroup):
        self.configuration = configuration

    def generate(self):

        shortcut_str = self._generate_i3_shortcuts_string()
        i3_config_path = os.path.join(os.environ["HOME"], ".config/i3")
        if not os.path.exists(f"{i3_config_path}/config_part1"):
            raise Exception("Cannot find part 1 of i3 configuration")

        write_file(f"{i3_config_path}/config_part2", shortcut_str)

        shell.run(f"cat {i3_config_path}/config_part1 > {i3_config_path}/config")
        shell.run(f"cat {i3_config_path}/config_part2 >> {i3_config_path}/config")

    def _generate_i3_shortcuts_string(self) -> str:
        """Generates a single string with all exported shortcuts"""
        result = "#automatically generated from now on\n"

        for key, content in list(self.configuration.commands.items()):
            if type(content) is dict and "i3_shortcut" in content:
                logging.info(f"Generating shortcut for {key}")
                identifier = self._generate_identifier(key)
                cmd = f'python_search run_key "{identifier}" --force_gui_mode=1 --from_shortcut=1'
                result = f"{result}bindsym {content['i3_shortcut']} exec {cmd}\n"

        return result

    def _generate_identifier(self, string):
        """
        strip the string from all special characters lefting only [A-B-09]
        """
        result = "".join(e for e in string if e.isalnum())
        result = result.lower()

        return result



import logging
import os


class FeaturesSupport:
    """
    Enables us to have the core behaviour working without the latest in development features.[

    """

    DYNAMIC_RANKING = "dynamic_ranking"

    DEFAULT_SUPPORT = {
        # if the ml ranking should be used or not
        # if user history or is supported
        DYNAMIC_RANKING: False,
        # turn on if you have a redis instance to improve the ranking
        "redis": False,
        # turn to true if you want data to be collected
        "event_tracking": False,
    }

    @staticmethod
    def default():
        return FeaturesSupport(FeaturesSupport.DEFAULT_SUPPORT)

    def __init__(self, config: dict):
        self.supported_config = {**FeaturesSupport.DEFAULT_SUPPORT, **config}

    def is_enabled(self, feature_name) -> bool:
        if feature_name not in self.supported_config:
            raise Exception(f"Feature {feature_name} not configured")

        return self.supported_config[feature_name]

    def is_dynamic_ranking_supported(self):
        return self.is_enabled("dynamic_ranking")

    def is_redis_supported(self):
        return self.is_enabled("redis")


class FeatureToggle:
    """
    A simple feature toggle abstraction that allows one to turn on/off features for their environment without touching code.
    """

    def __init__(self):
        self.BASE_PATH = f"{os.getenv('HOME')}/.PythonSearch/features"

    def enable(self, feature_name: str):
        os.system(f"touch {self.BASE_PATH}/{feature_name}")

    def disable(self, feature_name: str):
        os.system(f"rm {self.BASE_PATH}/{feature_name}")

    def is_enabled(self, feature_name: str) -> bool:
        result = 0 == os.system(f" test -f {self.BASE_PATH}/{feature_name}")

        if result:
            logging.debug(f"Feature toggle {feature_name} is enabled")
        else:
            logging.debug(f"Feature toggle {feature_name} is disabled")

        return result



#!/usr/bin/env python3
import logging
import subprocess

from python_search.entries_group import EntriesGroup


class Gnome:
    def __init__(self, configuration: EntriesGroup):
        self.configuration = configuration

    def generate(self):
        print("Generating gnome shortctus")
        self._reset()

        shortcut_found = False
        for key, content in list(self.configuration.commands.items()):
            if type(content) is dict and "gnome_shortcut" in content:
                logging.info(f"Generating shortcut for {key}")
                identifier = self._generate_identifier(key)
                cmd = f'python_search run_key "{identifier}" --force_gui_mode=1 --from_shortcut=1'
                self.generate_shortcut(identifier, cmd, content["gnome_shortcut"])
                shortcut_found = True

        if not shortcut_found:
            print("No shorcut found for gnome")

    def _reset(self):
        """reset existing shortcuts, necessary only for gnome"""

        from python_search.interpreter.cmd import CmdInterpreter

        CmdInterpreter(
            {
                "cmd": "gsettings reset-recursively org.gnome.settings-daemon.plugins.media-keys"
            }
        )

    def generate_shortcut(self, name: str, command: str, binding: str):
        """
        Super key:                 <Super>
        Control key:               <Primary> or <Control>
        Alt key:                   <Alt>
        Shift key:                 <Shift>
        numbers:                   1 (just the number)
        Spacebar:                  space
        Slash key:                 slash
        Asterisk key:              asterisk (so it would need `<Shift>` as well)
        Ampersand key:             ampersand (so it would need <Shift> as well)

        a few numpad keys:
        Numpad divide key (`/`):   KP_Divide
        Numpad multiply (Asterisk):KP_Multiply
        Numpad number key(s):      KP_1
        Numpad `-`:                KP_Subtract
        """

        # defining keys & strings to be used
        key = "org.gnome.settings-daemon.plugins.media-keys custom-keybindings"
        subkey1 = key.replace(" ", ".")[:-1] + ":"
        item_s = "/" + key.replace(" ", "/").replace(".", "/") + "/"
        firstname = "custom"
        # get the current list of custom shortcuts
        get = lambda cmd: subprocess.check_output(["/bin/bash", "-c", cmd]).decode(
            "utf-8"
        )
        array_str = get("gsettings get " + key)
        # in case the array was empty, remove the annotation hints
        command_result = array_str.lstrip("@as")
        current = eval(command_result)
        # make sure the additional keybinding mention is no duplicate
        n = 1
        while True:
            new = item_s + firstname + str(n) + "/"
            if new in current:
                n = n + 1
            else:
                break
        # add the new keybinding to the list
        current.append(new)
        # create the shortcut, set the name, command and shortcut key
        cmd0 = "gsettings set " + key + ' "' + str(current) + '"'
        cmd1 = "gsettings set " + subkey1 + new + " name '" + name + "'"
        cmd2 = "gsettings set " + subkey1 + new + " command '" + command + "'"
        cmd3 = "gsettings set " + subkey1 + new + " binding '" + binding + "'"

        for cmd in [cmd0, cmd1, cmd2, cmd3]:
            print(f"CMD executing: {cmd}")
            subprocess.call(["/bin/bash", "-c", cmd])

    def _generate_identifier(self, string):
        """
        strip the string from all special characters lefting only [A-B-09]
        """
        result = "".join(e for e in string if e.isalnum())
        result = result.lower()

        return result



import logging

from python_search.apps.browser import Browser
from python_search.exceptions import CommandDoNotMatchException
from python_search.interpreter.base import BaseInterpreter
from python_search.interpreter.cmd import CmdInterpreter


class UrlInterpreter(BaseInterpreter):
    def __init__(self, cmd, context=None):
        self.context = context

        if type(cmd) == str and UrlInterpreter.is_url(cmd):
            self.cmd = {"url": cmd}
            return

        if type(cmd) is dict and "url" in cmd:
            self.cmd = cmd
            return

        raise CommandDoNotMatchException(f"Not Valid URL command {cmd}")

    def interpret_default(self):
        logging.info(f'Processing as url: {self.cmd["url"]}')

        app_mode = self.cmd["app_mode"] if "app_mode" in self.cmd else False
        shell_cmd: str = Browser().open_cmd(self.cmd["url"], app_mode)

        final_cmd = self.cmd
        final_cmd["cmd"] = shell_cmd

        print(f"Final command={final_cmd}")
        return CmdInterpreter(final_cmd, self.context).interpret_default()

    def copiable_part(self):
        return self.cmd["url"]

    @staticmethod
    def is_url(url_candidate) -> bool:
        return url_candidate.startswith("http")



"""
Clients should depend on a configuration instance (config) rather than in the class,
the class should only be used for type annotation.
This way we can have multiple configs depending of the enviroment.
"""
import logging
import os
from typing import List, Optional

from python_search.entries_group import EntriesGroup
from python_search.environment import is_mac
from python_search.features import FeaturesSupport


class SearchRunConfiguration:
    NLP_PICKLED_EMBEDDINGS: str = f"{os.getenv('HOME')}/.python_search_nlp_embeddings"
    # editor used to edit the entries
    EDITOR = "vim"


class DataConfig:
    """
    Configuration with locations of data, models and names of files.
    """

    # output of the model
    base_data_folder = "/data/python_search"
    prediction_batch_location = base_data_folder + "/predict_input_lenght/latest"
    # a copy of the search run entries for the feature store
    entries_dump = base_data_folder + "/entries_dumped/latest"
    entries_dump_file = base_data_folder + "/entries_dumped/latest/000.parquet"
    cached_configuration = "/tmp/search_and_run_configuration_cached"
    MLFLOW_MODELS_PATH = f"{os.environ['HOME']}/projects/PySearchEntries/mlflow"
    BASELINE_EXPERIMENT_NAME = f"baseline_rank_v0"
    NEXT_ITEM_EXPERIMENT_NAME = f"next_item_v0"
    DATA_WAREHOUSE_FOLDER = base_data_folder + "/data_warehouse"
    SEARCH_RUNS_PERFORMED_FOLDER = (
        base_data_folder + "/data_warehouse/dataframes/SearchRunPerformed"
    )


# @todo do not depend on this config directly rather depend on the base configuration
config = SearchRunConfiguration()


class KafkaConfig:
    default_port: str = "9092"
    host: str = f"127.0.0.1:{default_port}"


class RedisConfig:
    host = "localhost"
    port = 6379 if is_mac() else 6378


class PythonSearchConfiguration(EntriesGroup):
    """
    The main configuration of Python Search
    Everything to customize about the application is configurable via code through this class
    """

    APPLICATION_TITLE = "SearchPythonSearch"
    commands: dict
    simple_gui_theme = "SystemDefault1"
    simple_gui_font_size = 14

    def __init__(
        self,
        *,
        entries: Optional[dict] = None,
        entries_groups: Optional[List[EntriesGroup]] = None,
        supported_features: Optional[FeaturesSupport] = None,
    ):
        if entries:
            self.commands = entries

        if entries_groups:
            self.aggregate_commands(entries_groups)

        if supported_features:
            self.supported_features: FeaturesSupport = supported_features
        else:
            self.supported_features: FeaturesSupport = FeaturesSupport.default()


class ConfigurationLoader:
    """
    Loads the application from the environment.py
    """

    def load_config(self) -> PythonSearchConfiguration:

        folder = self.get_project_root()
        config_location = f"{folder}/entries_main.py"

        if not os.path.exists(config_location):
            raise Exception(f"Could not find config file {config_location}")

        import sys

        sys.path.insert(0, folder)
        from entries_main import config

        return config

    def get_project_root(self):
        env_name = "PS_ENTRIES_HOME"
        current_project_location = (
            os.environ["HOME"] + "/.config/python_search/current_project"
        )

        folder = None

        if env_name in os.environ:
            logging.debug(
                f"Env exists and takes precedence: {env_name}={os.environ[env_name]}"
            )
            folder = os.environ[env_name]

        if os.path.isfile(current_project_location):
            with open(current_project_location) as f:
                folder = f.readlines()[0].strip()

        if not folder:
            raise Exception(
                f"Either {current_project_location} or {env_name} must be set to find entries"
            )
        return folder

    def load_entries(self):
        config = self.load_config()
        return config.commands

    def reload(self):
        import importlib

        import entries_main

        importlib.reload(entries_main)
        config = entries_main.config

        return config



import matplotlib.pyplot as plt
import numpy as np
import xgboost as xgb
from sklearn.metrics import mean_absolute_error
from sklearn.model_selection import train_test_split

from python_search.config import DataConfig
from python_search.ranking.next_item_predictor.offline_evaluation import \
    OfflineEvaluation
from python_search.ranking.next_item_predictor.transform import Transform


class TrainXGBoost:
    def train_and_log(self, dataset):
        """
        train the model and log it to MLFlow
        """
        import mlflow

        mlflow.set_tracking_uri(f"file:{DataConfig.MLFLOW_MODELS_PATH}")
        # this creates a new experiment
        mlflow.set_experiment(DataConfig.NEXT_ITEM_EXPERIMENT_NAME)
        mlflow.xgboost.autolog()

        with mlflow.start_run():
            model, offline_evaluation = self.train(dataset)
            # mlflow.log_params(metrics)
            mlflow.log_params(offline_evaluation)

        return model, offline_evaluation

    def train(self, dataset):

        X_train, X_test, Y_train, Y_test = self._transform_and_split(dataset)

        X_test_p = X_test
        X_test = np.delete(X_test, 0, axis=1)
        X_train = np.delete(X_train, 0, axis=1)

        model = xgb.XGBRegressor(
            tree_method="hist",
            eval_metric=mean_absolute_error,
            objective="reg:squarederror",
            eval_set=[(X_train, Y_train), (X_test, Y_test)],
            booster="gbtree",
            verbose=True,
        )

        eval_set = [(X_train, Y_train), (X_test, Y_test)]
        model.fit(X_train, Y_train, eval_set=eval_set)

        self._evaluate_metrics(model)
        offline_evaluation = OfflineEvaluation().run(model, dataset, X_test_p)
        print(offline_evaluation)

        return model, offline_evaluation

    def _transform_and_split(self, dataset):

        X, Y = Transform().transform_train(dataset)

        X_train, X_test, Y_train, Y_test = train_test_split(
            X, Y, test_size=0.10, random_state=42
        )
        X_test = np.where(np.isnan(X_test), 0.5, X_test)
        Y_test = np.where(np.isnan(Y_test), 0.5, Y_test)
        Y_train = np.where(np.isnan(Y_train), 0.5, Y_train)
        X_train = np.where(np.isnan(X_train), 0.5, X_train)

        return X_train, X_test, Y_train, Y_test

    def _evaluate_metrics(self, model, display_metrics=False):

        results = model.evals_result()
        epochs = len(results["validation_0"]["rmse"])
        x_axis = range(0, epochs)
        fig, ax = plt.subplots()
        ax.plot(x_axis, results["validation_0"]["rmse"], label="Train")
        ax.plot(x_axis, results["validation_1"]["rmse"], label="Test")
        ax.legend()
        plt.ylabel("rmse")
        plt.title("XGBoost_training rmse")
        if display_metrics:
            plt.show()

        results = model.evals_result()
        epochs = len(results["validation_0"]["mean_absolute_error"])
        x_axis = range(0, epochs)
        fig, ax = plt.subplots()
        ax.plot(x_axis, results["validation_0"]["mean_absolute_error"], label="Train")
        ax.plot(x_axis, results["validation_1"]["mean_absolute_error"], label="Test")
        ax.legend()
        plt.ylabel("mae")
        plt.title("XGBoost_training mae")
        if display_metrics:
            plt.show()



import logging
from typing import Optional

from grimoire.shell import shell
from grimoire.string import remove_special_chars

# @ todo remove this dependencies on grimoire
from python_search.apps.notification_ui import send_notification
from python_search.apps.terminal import Terminal
from python_search.context import Context
from python_search.exceptions import CommandDoNotMatchException
from python_search.interpreter.base import BaseInterpreter

# @todo find a better name
WRAP_IN_TERMINAL = "new-window-non-cli"


class CmdInterpreter(BaseInterpreter):
    """
    Represents a bash command entry. It can be used anywhere to run bash commends.
    """

    def __init__(self, cmd, context: Optional[Context] = None):
        """ """
        self.context = context

        if type(cmd) == str:
            self.cmd = {WRAP_IN_TERMINAL: True, "cmd": cmd}
            return

        if WRAP_IN_TERMINAL in cmd and "cmd" in cmd:
            self.cmd = cmd
            return

        if "cli_cmd" in cmd:
            self.cmd = cmd
            self.cmd["cmd"] = cmd["cli_cmd"]
            self.cmd[WRAP_IN_TERMINAL] = True
            return

        if "cmd" in cmd:
            self.cmd = cmd
            return

        raise CommandDoNotMatchException.not_valid_command(self, cmd)

    def interpret_default(self):
        if self.try_to_focus():
            return

        cmd = self.apply_directory(self.cmd["cmd"])

        if "high_priority" in self.cmd:
            cmd = f"nice -19 {cmd}"

        if "directory" in self.cmd:
            cmd = f'cd {self.cmd["directory"]} && {cmd}'

        if "tmux" in self.cmd:
            cmd = f'tmux new -s "{self._get_window_title()}" {cmd} '

        if WRAP_IN_TERMINAL in self.cmd:
            cmd = self._try_to_wrap_in_terminal(cmd)

        print(f"Command to run: {cmd}")
        result = self._execute(cmd)
        print(f"Result finished: {result}")
        return self.return_result(result)

    def _try_to_wrap_in_terminal(self, cmd):
        if WRAP_IN_TERMINAL not in self.cmd:
            return cmd

        logging.info("Running it in a new terminal")

        hold_terminal = False if "not_hold_terminal" in self.cmd else True
        cmd = Terminal().wrap_cmd_into_terminal(
            cmd,
            title=self._get_window_title(),
            hold_terminal_open_on_end=hold_terminal,
        )
        logging.info(f"Command to run: {cmd}")

        return cmd

    def _get_window_title(self):
        if "window_title" in self.cmd:
            return self.cmd["window_title"]

        title = self.cmd["cmd"]
        if "focus_match" in self.cmd:
            title = self.cmd["focus_match"]

        return remove_special_chars(title, [" "])

    def _execute(self, cmd):
        logging.info(f"To run: {cmd}")

        hold_terminal = False if "not_hold_terminal" in self.cmd else True
        if (
            self.context
            and self.context.is_group_command()
            and not self.context.should_execute_sequentially()
        ) or not hold_terminal:
            return shell.run_command_no_wait(cmd)

        if self.context and self.context.should_execute_sequentially():
            return shell.run_with_result(cmd)

        import subprocess

        p = subprocess.Popen(
            cmd,
            shell=True,
            stdin=None,
            stdout=None,
            stderr=None,
            close_fds=True,
        )

        return {"pid": p.pid}

    def return_result(self, result):
        if "notify-result" in self.cmd:
            send_notification(result)

        return result

    def copiable_part(self):
        return self.cmd["cmd"]



# !/usr/bin/env python import fire
import PySimpleGUI as sg


class Cli:
    def run(self):
        sg.Window(title="Hello World", layout=[[]], margins=(100, 50)).read()
        return "hello world"


if __name__ == "__main__":
    fire.Fire(Cli)



import logging

from python_search.apps.window_manager import WindowManager
from python_search.config import ConfigurationLoader, PythonSearchConfiguration
from python_search.entry_runner import EntryRunner
from python_search.environment import is_mac
from python_search.search_ui.preview import Preview


class PythonSearchCli:
    """
    The command line application, entry point of the program.

    Try to avoid adding direct commands, prefer instead to add objects as parts of functions
    """

    configuration: PythonSearchConfiguration

    @staticmethod
    def setup_from_config(config: PythonSearchConfiguration):
        """Initialized the cli with the main configuration object"""
        try:
            instance = PythonSearchCli(config)
            import fire

            fire.Fire(instance)
        except BaseException as e:
            _error_handler(e)

    @staticmethod
    def new_project(project_name: str):
        """Create a new project in the current directory with the given name"""
        from python_search.init_project import Project

        Project().new(project_name)

    def __init__(self, configuration: PythonSearchConfiguration = None):
        """
        Keep this constructor small and import dependencies inside the functions
        so they keep being fast
        """
        if not configuration:
            logging.debug("No configuration provided, using default")
            try:
                configuration = ConfigurationLoader().load_config()
            except Exception as e:
                print(f"Did not find any config to load, error: {e}")
                import traceback

                print(traceback.format_exc())

                return

        self.configuration = configuration
        self.run_key = EntryRunner(self.configuration).run

    def search(self):
        """
        Opens the Search UI. Main entrypoint of the application
        """
        from python_search.search_ui.search import Search

        Search(self.configuration).run()

    def copy_entry_content(self, key: str):
        """
        Copies the content of the provided key to the clipboard.
        Used by fzf to provide Ctrl-c functionality.
        """
        from python_search.interpreter.interpreter_matcher import \
            InterpreterMatcher

        InterpreterMatcher.build_instance(self.configuration).clipboard(key)

    def copy_key_only(self, key_str: str):
        """
        Copies to clipboard the key
        """
        from python_search.apps.clipboard import Clipboard

        Clipboard().set_content(key_str.split(":")[0])

    def edit_key(self, key):
        from python_search.entry_capture.edit_content import EditKey

        return EditKey(self.configuration).edit_key(key, dry_run=False)

    def search_edit(self, key=None):
        from python_search.entry_capture.edit_content import EditKey

        return EditKey(self.configuration).search_entries_directory(key)

    def edit_main(self):
        """Edit the main script"""
        from python_search.entry_capture.edit_content import EditKey

        return EditKey(self.configuration).edit_default()

    def register_clipboard(self):
        from python_search.entry_capture.register_new import RegisterNew

        return RegisterNew(self.configuration).launch_ui()

    def register_new(self):
        from python_search.entry_capture.register_new import RegisterNew

        return RegisterNew(self.configuration)

    def shortcut_generator(self):
        """Generate shorcuts for all environments"""
        from python_search.shortcut.generator import ShortcutGenerator

        return ShortcutGenerator(self.configuration)

    def ranking(self):
        from python_search.ranking.ranking import RankingGenerator

        return RankingGenerator(self.configuration)

    def consumers(self):
        """
        Provides access to the event consumers
        """
        from python_search.events.latest_used_entries import LatestUsedEntries

        class Consumers:
            def latest_used_entries(self):
                LatestUsedEntries().consume()

        return Consumers()

    def features(self):
        """Feature toggle system"""
        from python_search.features import FeatureToggle

        return FeatureToggle()

    def _utils(self):
        """Here commands that are small topics and dont fit the rest"""

        class Utils:
            def __init__(self, configuration):
                self.configuration = configuration

            def hide_launcher(self):
                """hide the search launcher -i2 specific"""
                if is_mac():
                    import os

                    os.system(
                        """osascript -e 'tell application "System Events" to keystroke "H" using {command down}'"""
                    )
                WindowManager.load_from_environment().hide_window(
                    self.configuration.APPLICATION_TITLE
                )

        return Utils(self.configuration)

    def _preview_entry(self, entry_text: str):
        """
        Recieves entries from fzf and show them formatted for the preview window
        """
        Preview().display(entry_text)

    def google_it(self, query):
        from python_search.interpreter.url import UrlInterpreter

        UrlInterpreter(
            {
                "url": f"http://www.google.com/search?q={query}",
            }
        ).interpret_default()

    def _infra_report(self):
        from python_search.infrastructure.report import Report

        return Report()


def _error_handler(e):
    from python_search.observability.logger import initialize_systemd_logging

    logging = initialize_systemd_logging()
    import sys
    import traceback

    exc_info = sys.exc_info()
    logging.warning(
        f"Unhandled exception: {e}".join(traceback.format_exception(*exc_info))
    )

    raise e


def _run_key_bin():
    import fire

    fire.Fire(PythonSearchCli().run_key)


def main():
    import fire

    fire.Fire(PythonSearchCli)



from dataclasses import dataclass
from typing import List

import fire

from python_search.config import ConfigurationLoader


@dataclass
class EntryData:
    """
    Entry data schema

    """

    key: str
    value: str
    type: str
    tags: List[str]


class EntryCaptureGUI:
    TAGS = [
        "German",
        "Reminder",
        "Politics",
        "Self",
        "Others",
        "Career",
        "Commitment",
        "StrategicDecision",
        "Linux",
        "Mac",
    ]

    def launch(
        self,
        title: str = "Capture Entry",
        default_key="",
        default_content: str = "",
        serialize_output=False,
        default_type="Snippet",
    ) -> EntryData:
        """
        Launch the data capture GUI.
        """
        import PySimpleGUI as sg

        config = ConfigurationLoader().load_config()
        sg.theme(config.simple_gui_theme)
        font_size = config.simple_gui_font_size

        content_input = sg.Input(
            key="content",
            default_text=default_content,
            expand_x=True,
            expand_y=True,
        )
        layout = [
            [sg.Text("Entry content")],
            [content_input],
            [sg.Text("Descriptive key name")],
            [
                sg.Input(
                    key="key", default_text=default_key, expand_x=True, expand_y=True
                )
            ],
            [sg.Text("Type")],
            [
                sg.Combo(
                    [
                        "Snippet",
                        "Cmd",
                        "Url",
                        "File",
                        "Anonymous",
                    ],
                    key="type",
                    default_value=default_type,
                )
            ],
            [sg.Text("Tags")],
            [sg.Checkbox(tag, key=tag, default=False) for tag in EntryCaptureGUI.TAGS],
            [sg.Button("Write", key="write")],
        ]

        window = sg.Window(
            title,
            layout,
            font=("Helvetica", font_size),
            alpha_channel=0.9,
            finalize=True,
        )

        # workaround for mac bug
        window.read(timeout=100)
        window.set_alpha(1.0)
        content_input.update(select=True)

        window["key"].bind("<Return>", "_Enter")
        window["content"].bind("<Return>", "_Enter")
        window["type"].bind("<Return>", "_Enter")
        window["key"].bind("<Escape>", "_Esc")
        window["content"].bind("<Escape>", "_Esc")
        window["type"].bind("<Escape>", "_Esc")

        while True:
            event, values = window.read()
            if event and (event == "write" or event.endswith("_Enter")):
                break
            if event == sg.WINDOW_CLOSED or event.endswith("_Esc"):
                raise Exception("Quitting window")

        window.close()
        print("values", values)

        selected_tags = []
        for key, value in values.items():
            if key in EntryCaptureGUI.TAGS and value == True:
                selected_tags.append(key)

        result = EntryData(
            values["key"], values["content"], values["type"], selected_tags
        )

        if serialize_output:
            result = result.__dict__

        return result


if __name__ == "__main__":
    fire.Fire(EntryCaptureGUI().launch)



from __future__ import annotations

import datetime
import time
from typing import Tuple

from grimoire.event_sourcing.message import MessageBroker
from grimoire.string import emptish

from python_search.apps.clipboard import Clipboard
from python_search.entry_capture.entry_inserter import EntryInserter
from python_search.entry_capture.gui import EntryCaptureGUI, EntryData
from python_search.exceptions import RegisterNewException
from python_search.interpreter.base import BaseInterpreter
from python_search.interpreter.file import FileInterpreter
from python_search.interpreter.interpreter_matcher import InterpreterMatcher
from python_search.interpreter.url import UrlInterpreter


class RegisterNew:
    """
    Responsible for registering new entries in your catalog
    """

    def __init__(self, configuration):
        self.configuration = configuration
        self.message_broker = MessageBroker("search_run_register_new")
        self.entry_inserter = EntryInserter(configuration)

    def launch_ui(self, default_type=None, default_key=None, default_content=None):
        """
        Create a new inferred entry based on the clipboard content
        """

        if not default_content:
            default_content = Clipboard().get_content()

        if not default_type:
            default_type = self._infer_default_type(default_content)

        entry_data: EntryData = EntryCaptureGUI().launch(
            "New Entry Details",
            default_content=default_content,
            default_key=default_key,
            default_type=default_type,
        )
        interpreter: BaseInterpreter = InterpreterMatcher.build_instance(
            self.configuration
        ).get_interpreter_from_type(entry_data.type)

        dict_entry = interpreter(entry_data.value).to_dict()
        if entry_data.tags:
            dict_entry["tags"] = entry_data.tags

        self.entry_inserter.insert(entry_data.key, dict_entry)

    def _infer_default_type(self, content):
        if UrlInterpreter.is_url(content):
            return "Url"

        if FileInterpreter.file_exists(content):
            return "File"

        return "Cmd"

    def anonymous_snippet(self):
        """
        Create an anonymous snippet entry
        """

        now = datetime.datetime.now()
        key = f"no key {now.strftime('%Y %M %d %H %M %S')}"

        self.launch_ui(default_key=key, default_type="Snippet")

    def german_from_text(self, key):
        """
        Register german workds you dont know by saving them to the clipboard and storing in python search
        """

        if emptish(key):
            raise RegisterNewException.empty_content()

        from python_search.interpreter.url import UrlInterpreter

        cmd = {
            "url": f"https://translate.google.com/?sl=de&tl=en&text={key}&op=translate"
        }
        UrlInterpreter(cmd).interpret_default()
        time.sleep(1)

        from python_search.apps.collect_input import CollectInput

        meaning = CollectInput().launch(f"Please type the meaning of ({key})")

        if emptish(meaning):
            raise RegisterNewException.empty_content()

        as_dict = {
            "snippet": meaning,
            "language": "German",
        }

        self.entry_inserter.insert(key, as_dict)



import sys

import fire

from python_search.apps.clipboard import Clipboard


class CollectInput:
    """
    Ask the user for input and return the entered data
    """

    def launch(self, name="Enter Data", prefill_with_clipboard: bool = False):
        """
        Launch the data capture GUI.
        """
        import PySimpleGUI as sg

        default_text = ""
        if prefill_with_clipboard:
            default_text = Clipboard().get_content()

        font_size = 12
        sg.theme("SystemDefault1")

        input_field = sg.Input(
            key="content", default_text=default_text, expand_x=True, expand_y=True
        )

        layout = [
            [input_field],
            [sg.Button("Continue", key="write")],
        ]

        window = sg.Window(
            name,
            layout,
            finalize=True,
            font=("Helvetica", font_size),
            alpha_channel=0.9,
        )

        # workaround for mac bug
        window.read(timeout=100)
        if default_text != "":
            input_field.update(select=True)
        window.set_alpha(1.0)

        window["content"].bind("<Return>", "_Enter")
        window["content"].bind("<Escape>", "_Esc")

        while True:
            event, values = window.read()

            if event and (event == "write" or event.endswith("_Enter")):
                break
            if event == sg.WINDOW_CLOSED or event.endswith("_Esc"):
                sys.exit(1)

        window.close()

        return values["content"]


def main():
    fire.Fire(CollectInput().launch)


if __name__ == "__main__":
    main()



from typing import List


def generate_acronyms(name: str) -> List[str]:
    return [generate_acronym(name)]


def generate_acronym(name: str) -> str:
    words = name.split(" ")
    acronyms = [word[0] for word in words]
    return "".join(acronyms)



from __future__ import annotations

import copy

from python_search.events.latest_used_entries import LatestUsedEntries
from python_search.ranking.entry_embeddings import (EmbeddingSerialization,
                                                    RedisEmbeddingsReader,
                                                    RedisEmbeddingsWriter)


class InferenceEmbeddingsLoader:
    """
    @Todo, unify embeddings loading for inference but not training
    """

    def __init__(self, all_keys):
        self.all_keys = copy.copy(list(all_keys))
        self.latest_used_entries = LatestUsedEntries()
        self.embedding_mapping = RedisEmbeddingsReader().load(self.all_keys)

    def get_embedding_from_key(self, key: str):
        """
        Return an embedding based on the keys
        """
        if not key in self.embedding_mapping or self.embedding_mapping[key] is None:
            print(
                f"The embedding for ({key}) is empty in redis. Syncing the missing keys"
            )
            RedisEmbeddingsWriter().sync_missing()
            self.embedding_mapping = RedisEmbeddingsReader().load(self.all_keys)

            if self.embedding_mapping[key] is None:
                raise Exception(
                    f"After trying to sync embedding  for key '{key}' is still empty"
                )

        return EmbeddingSerialization.read(self.embedding_mapping[key])

    def get_recent_key_with_embedding(self, second_recent=False) -> str:
        """
        Look into the recently used keys and return the most recent for which there are embeddings
        """
        iterator = self.latest_used_entries.get_latest_used_keys()
        print("On get_recent_key all keys size: " + str(len(self.all_keys)))
        print("Number of latest used keys: " + str(len(iterator)))

        print("Mapping size: " + str(len(self.embedding_mapping)))
        first_found = False
        for previous_key in iterator:
            if previous_key not in self.embedding_mapping:
                print(f"Key {previous_key} not found in mapping")
                continue
            if not self.embedding_mapping[previous_key]:
                print("Key found but no content in: ", previous_key)
                continue

            if second_recent and not first_found:
                first_found = True
                continue

            return previous_key

        print_mapping = False
        extra_message = ""
        if print_mapping:
            extra_message = "Existing keys: " + str(self.embedding_mapping.keys())

        raise Exception(f"Could not find a recent key with embeddings" + extra_message)



from python_search.search_ui.fzf_terminal import FzfInTerminal


class Search:
    """
    Opens search with all entries
    """

    def __init__(self, configuration):
        self.configuration = configuration

    def run(self):
        return FzfInTerminal.build_search_ui(self.configuration).run()



from __future__ import annotations

import shutil
from datetime import datetime

from python_search.config import PythonSearchConfiguration


class EntryInserter:
    """
    Class responsible to the actual writing of the new entry
    """

    NEW_ENTRIES_STRING = "# NEW_ENTRIES_HERE"

    def __init__(self, configuration: PythonSearchConfiguration):
        self.configuration = configuration
        self.file_to_append = self.configuration.get_project_root() + "/entries_main.py"

    def insert(self, key: str, entry: dict, enable_shortcuts_generation=False):

        entry["created_at"] = datetime.now().isoformat()

        row_entry = str(entry)
        line_to_add = f"    '{key}': {row_entry},"
        self._append_entry(line_to_add)

        if self.configuration.supported_features.is_enabled("event_tracking"):
            from python_search.events.events import SearchRunPerformed
            from python_search.events.producer import EventProducer

            EventProducer().send_object(
                SearchRunPerformed(key=key, query_input="", shortcut=False)
            )

        from python_search.apps.notification_ui import send_notification

        send_notification(f"Entry {row_entry} inserted successfully")

        # refresh the configuration
        if enable_shortcuts_generation:
            from python_search.interpreter.cmd import CmdInterpreter

            CmdInterpreter({"cmd": "python_search generate_shortcuts"})

    def _append_entry(self, line_to_add: str):
        """
        This script does the following:
            Copies the main file,
            add the new entry to it
            Compile to see if it is still valid python
            If so, then replaces it
        """
        copy_file = self.file_to_append + "cpy"

        shutil.copyfile(self.file_to_append, copy_file)

        with open(copy_file, "w") as out, open(self.file_to_append) as source_file:
            for line in source_file:
                out.write(line)
                if self.NEW_ENTRIES_STRING in line:
                    # insert text.
                    print(f"Writing line: {line_to_add}")
                    out.write(line_to_add + "\n")

        # compile and make sure the file is a valid python
        import os

        if os.system(f"python3 -m compileall -q {copy_file}") != 0:
            message = "Copy of file does not compile so wont proceed replacing!"
            from python_search.apps.notification_ui import send_notification

            send_notification(message)
            raise Exception(message)

        shutil.move(copy_file, self.file_to_append)



import logging
import math
import os.path
import sys
from typing import Optional

import pyspark.sql.functions as F
from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.functions import struct, udf
from pyspark.sql.types import FloatType
from pyspark.sql.window import Window

from python_search.datasets.searchesperformed import SearchesPerformed
from python_search.infrastructure.performance import timeit


class TrainingDataset:
    """
    Builds the dataset ready for training
    """

    columns = (
        "key",
        "previous_key",
        "previous_previous_key",
        "month",
        "hour",
        "label",
        "entry_number",
    )
    _DATASET_CACHE_FILE = "/tmp/dataset"

    def __init__(self):
        self._spark = SparkSession.builder.getOrCreate()
        logging.basicConfig(
            level=logging.INFO, handlers=[logging.StreamHandler(sys.stdout)]
        ),
        self._dataframe = None

    @timeit
    def build(self, use_cache=False, write_cache=True) -> DataFrame:
        """
        When cache is enabled, writes a parquet in a temporary file
        """
        if use_cache:
            cache = self._read_cache()
            if cache:
                return cache

        dataset_with_aggregations = self._prepare_dataset_with_aggregations()
        dataset = self._add_label_and_cleanup(dataset_with_aggregations)

        logging.info("TrainingDataset ready, writing it to disk")
        if write_cache:
            self._write_cache(dataset)

        logging.info("Printing a sample of the dataset")
        dataset.show(10)

        self._dataframe = dataset
        return dataset

    def _prepare_dataset_with_aggregations(self):
        search_performed_df = self._load_base()
        search_performed_df_filtered = self._filter_blacklisted(search_performed_df)
        all_dimensions = self._add_all_features(search_performed_df_filtered)
        base_dataset = self._base_dataset(all_dimensions)
        dataset_with_aggregations = self._compute_aggregations(
            all_dimensions, base_dataset
        )

        return dataset_with_aggregations

    def _add_label_and_cleanup(self, all_features: DataFrame) -> DataFrame:
        """
        Remove all columns which the purpose is to calculate the label
        """

        logging.info("Adding label")

        def label_formula(row):
            return (
                math.log(row["times_3"])
                + math.log(row["times_2"] * 0.5)
                + math.log(row["global_pair"] * 0.01)
            )

        udf_f = udf(label_formula, FloatType())
        with_label = all_features.withColumn(
            "label", udf_f(struct([all_features[x] for x in all_features.columns]))
        )

        with_label.cache()
        # normalize label
        max_label = (
            with_label.agg({"label": "max"})
            .withColumnRenamed("max(label)", "max_label")
            .collect()[0]
            .max_label
        )
        max_label = float(max_label)
        print(f"Max value for label: {max_label}")

        min_label = (
            with_label.agg({"label": "min"})
            .withColumnRenamed("min(label)", "min_label")
            .collect()[0]
            .min_label
        )
        min_label = float(min_label)
        print(f"Min value for label: {min_label}")

        normalized = with_label.withColumn(
            "label_normalized", (F.col("label") - min_label) / (max_label - min_label)
        )

        print("Replacing the label colum for the normalized version")
        normalized = normalized.withColumnRenamed("label", "label_original")
        result = normalized.withColumnRenamed("label_normalized", "label")

        final_result = result.select(*self.columns)

        print("Schema of final dataframe")
        final_result.printSchema()

        return final_result

    def _compute_aggregations(self, all_dimensions, base_features) -> DataFrame:
        """
        Adds aggregations of the data that supports the label formula
        Args:
            all_dimensions:
            base_features:

        Returns:

        """

        # adds performance in the dimension of the triple key,previous key,previous previous key
        grouped3 = (
            all_dimensions.groupBy(
                "month", "hour", "key", "previous_key", "previous_previous_key"
            )
            .agg(F.count("*").alias("times_3"))
            .sort("times_3", ascending=False)
        )
        base_new = base_features.join(
            grouped3,
            on=["month", "hour", "key", "previous_key", "previous_previous_key"],
        )

        # adds performance in the time dimension of the pair key-previous key
        grouped2 = (
            all_dimensions.groupBy("month", "hour", "key", "previous_key")
            .agg(F.count("*").alias("times_2"))
            .sort("times_2", ascending=False)
        )
        base_new = base_new.join(grouped2, on=["month", "hour", "key", "previous_key"])

        # adds global performance of the pair
        global_pair = (
            all_dimensions.groupBy("key", "previous_key")
            .agg(F.count("*").alias("global_pair"))
            .sort("global_pair", ascending=False)
        )

        base_new = base_new.join(global_pair, on=["key", "previous_key"])

        return base_new

    def _base_dataset(self, all_dimensions) -> DataFrame:
        """
        This is the base dataset that will be send to train
        but before we will add some aggregations to it so we can generate the desired label
        Args:
            all_dimensions:

        Returns:

        """
        base_features = all_dimensions.select(
            "month", "hour", "key", "previous_key", "previous_previous_key"
        ).distinct()

        window = Window.orderBy(F.col("key"))

        # @todo verify if this number is correct auto-incremented and unique
        base_features = base_features.withColumn(
            "entry_number", F.row_number().over(window)
        )

        print("Base feature")
        base_features.show()

        return base_features

    def _load_base(self) -> DataFrame:
        return SearchesPerformed().load()

    def _add_all_features(self, df: DataFrame) -> DataFrame:
        """Return a dataframe with the hole features it will need but not aggregated"""
        logging.info("Loading searches performed")
        df_with_previous = self._join_with_previous(df)
        print("Add date dimensions")
        with_month = df_with_previous.withColumn("month", F.month("timestamp"))
        with_hour = with_month.withColumn("hour", F.hour("timestamp"))

        # keep only the necessary columns
        return with_hour.select(
            "month", "hour", "key", "previous_key", "previous_previous_key", "timestamp"
        )

    def _filter_blacklisted(self, df) -> DataFrame:
        # filter out too common keys
        EXCLUDED_ENTRIES = [
            "startsearchrunsearch",
            "search run search focus or open",
            "",
        ]
        return df.filter(~F.col("key").isin(EXCLUDED_ENTRIES))

    def __repr__(self):
        return self._dataframe.show(10)

    @timeit
    @timeit
    def _join_with_previous(self, df):
        """
        Adds the previou_key as an entry
        """
        # build pair dataset with label
        # add literal column
        search_performed_df_tmpcol = df.withColumn("tmp", F.lit("toremove"))
        window = Window.partitionBy("tmp").orderBy("timestamp")

        # add row number to the dataset
        search_performed_df_row_number = search_performed_df_tmpcol.withColumn(
            "row_number", F.row_number().over(window)
        ).sort("timestamp", ascending=False)

        # add previous key to the dataset
        search_performed_df_with_previous = search_performed_df_row_number.withColumn(
            "previous_key", F.lag("key", 1, None).over(window)
        )
        search_performed_df_with_previous_previous = (
            search_performed_df_with_previous.withColumn(
                "previous_previous_key", F.lag("key", 2, None).over(window)
            )
        )

        result = search_performed_df_with_previous_previous.sort(
            "timestamp", ascending=False
        )
        logging.debug("Showing it merged with 2 previous keys", result.show(10))

        return result

    def _write_cache(self, dataset) -> None:
        print("Writing cache dataset to disk")
        if os.path.exists(TrainingDataset._DATASET_CACHE_FILE):
            import shutil

            shutil.rmtree(TrainingDataset._DATASET_CACHE_FILE)

        dataset.write.parquet(TrainingDataset._DATASET_CACHE_FILE)

    def _read_cache(self) -> Optional[DataFrame]:
        if os.path.exists(TrainingDataset._DATASET_CACHE_FILE):
            print("Reading cache dataset")
            return self._spark.read.parquet(TrainingDataset._DATASET_CACHE_FILE)
        else:
            print("Cache does not exist, creating dataset")


if __name__ == "__main__":
    import fire

    fire.Fire(TrainingDataset)



from __future__ import annotations

import json
import os
from collections import namedtuple
from typing import List, Optional, Tuple

from python_search.acronyms import generate_acronyms
from python_search.config import PythonSearchConfiguration
from python_search.features import FeatureToggle
from python_search.infrastructure.performance import timeit
from python_search.infrastructure.redis import PythonSearchRedis
from python_search.observability.logger import logging


class RankingGenerator:
    """
    Generates the ranking for python search
    """

    ModelInfo = namedtuple("ModelInfo", "features label")
    _model_info = ModelInfo(["position", "key_lenght"], "input_lenght")

    def __init__(self, configuration: Optional[PythonSearchConfiguration] = None):
        self.configuration = configuration
        self.feature_toggle = FeatureToggle()
        self.model = None
        self.debug = os.getenv("DEBUG", False)

        if self.configuration.supported_features.is_redis_supported():
            self.redis_client = PythonSearchRedis.get_client()

        self.used_entries: List[Tuple[str, dict]] = []

        if self.feature_toggle.is_enabled("ranking_next"):
            from python_search.ranking.next_item_predictor.inference.inference import \
                Inference

            self.inference = Inference(self.configuration)

    def generate_with_caching(self):
        """
        Uses cached rank if available and only add new keys on top
        """
        self.generate(recompute_ranking=False)

    @timeit
    def generate(self, print_entries=True, print_weights=False):
        """
        Recomputes the rank and saves the results on the file to be read
        """

        self.entries: dict = self.configuration.commands
        # by default the rank is just in the order they are persisted in the file
        self.ranked_keys: List[str] = self.entries.keys()

        try:
            """Mutate self.ranked keys with the results"""
            self.ranked_keys = self.inference.get_ranking(print_weights=print_weights)
        except Exception as e:
            print(f"Inference failed with error {e} falling back to default ranking")

        """Populate the variable used_entries  with the results from redis"""
        self._fetch_latest_entries()
        result, only_list = self._merge_and_build_result()

        if not print_entries:
            return

        return self.print_entries(result)

    def _save_ranking_order_in_cache(self, ranking: List[str]):
        encoded_list = "|".join(ranking)
        self.redis_client.set("cache_ranking_result", encoded_list)

    def _can_load_from_cache(self):
        return (
            self.configuration.supported_features.is_redis_supported()
            and self.configuration.supported_features.is_dynamic_ranking_supported()
        )

    def _merge_and_build_result(self) -> List[Tuple[str, dict]]:
        """ "
        Merge the ranking with the latest entries and make it ready to be printed
        """
        result = []
        increment = 0
        final_key_list = []

        while self.used_entries:
            used_entry = self.used_entries.pop()
            key = used_entry[0]
            if key not in self.entries:
                # key not found in entries
                continue

            # sometimes there can be a bug of saving somethign other than dicts as entries
            if type(used_entry[1]) != dict:
                logging.warning(f"Entry is not a dict {used_entry[1]}")
                continue
            logging.debug(f"Increment: {increment}  with entry {used_entry}")
            final_key_list.append(used_entry[0])
            result.append((used_entry[0], {**used_entry[1], "recently_used": True}))
            increment += 1

        for key in self.ranked_keys:
            if key not in self.entries:
                # key not found in entries
                continue

            result.append((key, self.entries[key]))
            final_key_list.append(key)
            increment += 1

        # the result is the one to be returned, final_key_list is to be used in the cache
        return result, final_key_list

    def _fetch_latest_entries(self):
        """Populate the variable used_entries  with the results from redis"""
        self.used_entries: List[Tuple[str, dict]] = []
        if not self.configuration.supported_features.is_enabled(
            "redis"
        ) or not self.feature_toggle.is_enabled("ranking_latest_used"):
            if self.debug:
                print(f"Disabled latest entries")
            return

        self.used_entries = self.get_used_entries_from_redis(self.entries)
        # only use the latest 7 entries for the top of the ranking
        self.used_entries = self.used_entries[-7:]

        if self.debug:
            print(f"Used entries: {self.used_entries}")

    def get_used_entries_from_redis(self, entries) -> List[Tuple[str, dict]]:
        """
        returns a list of used entries to be placed on top of the ranking
        """
        used_entries = []
        latest_used = self._get_latest_used_keys()
        for used_key in latest_used:
            if used_key not in entries or used_key in used_entries:
                continue
            used_entries.append((used_key, entries[used_key]))
        # reverse the list given that we pop from the end
        used_entries.reverse()
        return used_entries

    def _get_latest_used_keys(self):
        from python_search.events.latest_used_entries import LatestUsedEntries

        return LatestUsedEntries().get_latest_used_keys()

    @timeit
    def print_entries(self, data: List[Tuple[str, dict]]) -> str:
        """Print results"""
        position = 1
        result = ""
        for name, content in data:
            name_clean = name.lower()
            try:
                content["key_name"] = name_clean
                content["position"] = position
                content["generated_acronyms"] = generate_acronyms(name)
                content_str = json.dumps(content, default=tuple, ensure_ascii=True)
            except BaseException as e:
                logging.debug(e)
                content_str = str(content)

            position = position + 1

            content_str = f"{name_clean}:" + content_str
            #  replaces all single quotes for double ones
            #  otherwise the json does not get rendered
            content_str = content_str.replace("'", '"')
            if os.getenv("ENABLE_TIME_IT"):
                # do not print if enable timeit is on
                continue
            result += content_str + "\n"
        return result


if __name__ == "__main__":
    import fire

    fire.Fire()



import os
import time
from datetime import timedelta
from functools import wraps


def timeit(method):
    """
    Abstraction to inspect functions performance.

    Example:

        @timeit
        def prepare_dataset()
            pass

        @timeit
        def train()
            pass

        then call your program with the variable enabled
        ENABLE_TIME_IT=True python run_pipeline.py

    """

    @wraps(method)
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = method(*args, **kwargs)
        end_time = time.time()
        if os.getenv("ENABLE_TIME_IT"):
            print(
                f"{method.__name__} => {timedelta(milliseconds = ((end_time-start_time) * 1000))} h:m:s"
            )

        return result

    return wrapper



""" Module responsible for the logic of editing entry files """
import logging
from typing import Optional

from grimoire import s
from grimoire.shell import shell

from python_search.apps.terminal import Terminal
from python_search.config import config
from python_search.interpreter.cmd import CmdInterpreter


class EditKey:
    """
    Set of commands to edit the entries
    """

    def __init__(self, configuration):
        self.configuration = configuration

    def edit_key(self, key, dry_run=False):
        """
        Edits the configuration files by searching the text
        """
        if not key:
            self._edit_default()
            return

        key = key.split(":")

        if not len(key):
            self.edit_default()
            return

        key = key[0]
        # needs to be case insensitive search
        cmd = f"ack -i '{key}' {self.configuration.get_project_root()} --py || true"

        logging.info(f"Command: {cmd}")
        result_shell = shell.run_with_result(cmd)
        if not result_shell:
            logging.info("Could not find match edit main file")
            self._edit_config(self.configuration.get_source_file(), dry_run)
            return

        file, line, *_ = result_shell.split(":")

        self._edit_config(file, line)

    def search_entries_directory(self, key=None):
        entry = {
            "cli_cmd": f"fzf_directory.sh {self.configuration.get_project_root()}",
            "directory": self.configuration.get_project_root(),
        }

        CmdInterpreter(entry).interpret_default()

    def edit_default(self):
        self._edit_config(self.configuration.get_project_root() + "/entries_main.py")

    def _edit_config(self, file_name: str, line: Optional[int] = 30, dry_run=False):
        """ "edit a configuration file given the name and line"""
        cmd: str = (
            f"MY_TITLE='GrimorieSearchRun' kitty {Terminal.GENERIC_TERMINAL_PARAMS} bash -c 'cd"
            f" {self.configuration.get_project_root()} "
            f"; {config.EDITOR} {file_name} +{line}' "
        )

        if dry_run:
            logging.info(f"Command to edit file: {cmd}")
            return

        s.run(cmd)



from python_search.config import DataConfig


class SearchesPerformed:
    """
    Poit of acess for the searches performed
    This is the place to keep the source of truth for the schema as well
    """

    columns = ["key", "query_input", "shortcut", "timestamp"]

    def __init__(self, spark=None):
        from pyspark.sql.session import SparkSession

        self.spark = spark if spark else SparkSession.builder.getOrCreate()

    def load(self):
        return self.spark.read.format("parquet").load(
            DataConfig.SEARCH_RUNS_PERFORMED_FOLDER
        )



import logging
import sys

from python_search.environment import is_mac

SYSLOG_IDENTIFIER = "python-search"


def initialize_logging():

    """Updates the global logging module"""
    handlers = [
        logging.StreamHandler(sys.stdout),
    ]
    logging.basicConfig(level=logging.INFO, handlers=handlers)

    return logging


def initialize_systemd_logging():
    return initialize_logging()



import json
import logging

from kafka import KafkaProducer

from python_search.config import KafkaConfig


class EventProducer:
    """Produce kafka messages"""

    def __init__(self):

        try:
            self.producer = KafkaProducer(
                bootstrap_servers=KafkaConfig.host,
                acks=0,
                max_block_ms=1000,
                request_timeout_ms=500,
            )
        except Exception as e:
            logging.info("Could not create kafka producer, not sending messages")
            self.producer = None

    def send_object(self, event_object):
        topic_name = type(event_object).__name__
        logging.info(f"Topic: {topic_name}")
        self.send(topic_name, event_object.__dict__)

    def send(self, topic_name: str, message: dict):
        if not self.producer:
            logging.warning("Could not create kakfa producer, not sending messages")
            return

        self.producer.send(topic_name, json.dumps(message).encode())
        self.producer.flush()



from __future__ import annotations

import os

import streamlit as st

from python_search.data_ui.training_page import load_training_page
from python_search.datasets.searchesperformed import SearchesPerformed

open_page = "results"

os.putenv("SPARK_LOCAL_IP", "localhost")
with st.sidebar:

    if st.button("Results evaluation"):
        open_page = "results"

    if st.button("Training Dataset"):
        open_page = "training"

    if st.button("Searches Performed Dataset"):
        open_page = "searches_performed_dataset"


if open_page == "training":
    load_training_page()

if open_page == "results":
    import python_search.data_ui.results_page as results_page

    results_page.load_results_page()

if open_page == "searches_performed_dataset":
    st.write("## Searches performed dataset")
    search_performed_df = SearchesPerformed().load()
    pdf = search_performed_df.toPandas()
    st.dataframe(pdf)



""" Centralize all exceptions definitions to help in the data discovery """


class MenuException(Exception):
    @staticmethod
    def given_empty_value():
        return MenuException("No option selected in search ui!")


class CommandDoNotMatchException(Exception):
    @staticmethod
    def not_valid_command(entity, cmd):
        return CommandDoNotMatchException(
            f"Not Valid {entity.__class__.__name__} command {cmd}"
        )


class RunException(Exception):
    @staticmethod
    def key_does_not_exist(key: str):
        return RunException(f"Key does not exist: {key}")


class RegisterNewException(Exception):
    @staticmethod
    def empty_content():
        return RegisterNewException(f"Will not register as content looks too small")



#!/usr/bin/env python3

import json
import logging
import sys

from kafka import KafkaConsumer

from python_search.config import KafkaConfig
from python_search.infrastructure.redis import PythonSearchRedis


class LatestUsedEntries:
    """
    Contains the logic to read and write the  latest used keys from redis
    These entries are then applied to the ranking.
    """

    MAX_PERSISTED_ITEMS = 1000
    NUMBER_OF_KEYS_TO_RETRIEVE = 100

    def __init__(self, key_name="latest_consumed_key"):
        logging.basicConfig(
            level=logging.INFO, handlers=[logging.StreamHandler(sys.stdout)]
        )

        self.redis_key_name = key_name
        self.redis_client = LatestUsedEntries.get_redis_client()

    @staticmethod
    def get_redis_client():
        return PythonSearchRedis.get_client()

    def get_latest_used_keys(self):
        """
        return a list of unike used keys ordered by the last time they were used
        the most recent in the top.
        """
        result = self.redis_client.lrange(
            LatestUsedEntries().redis_key_name,
            0,
            LatestUsedEntries.NUMBER_OF_KEYS_TO_RETRIEVE,
        )

        result = [x.decode() for x in result]
        result = list(dict.fromkeys(result))
        return result

    def write_last_used_key(self, value: str):
        self.redis_client.lpush(self.redis_key_name, value)
        self.redis_client.ltrim(self.redis_key_name, 0, self.MAX_PERSISTED_ITEMS)

    def consume(self):
        print("Starting kafka cosumer")
        consumer = KafkaConsumer(
            "SearchRunPerformed",
            bootstrap_servers=KafkaConfig.host,
            auto_offset_reset="earliest",
            enable_auto_commit=True,
            group_id="to_redis",
            value_deserializer=lambda x: json.loads(x.decode("utf-8")),
        )

        for message in consumer:
            key_executed = message.value["key"]
            if not message.value["shortcut"]:
                logging.info(f"THE key: {key_executed} will be persisted in redis")
                self.write_last_used_key(key_executed)
            else:
                logging.info(
                    f"THE key: {key_executed} will be skipped as it is a shortcut"
                )


if __name__ == "__main__":
    import fire

    fire.Fire(LatestUsedEntries)



""" Centralize all events definitions to help in the data discovery """
from dataclasses import dataclass

from pydantic import BaseModel


@dataclass
class SearchRunPerformed:
    """
    Main event of the application.
    Identifies a search being executed
    """

    # name of the entry matched
    key: str
    # for when a query was typed by the user
    query_input: str
    # for when it is started from a shortcut
    shortcut: str

    @staticmethod
    def get_schema():
        return "key string, query_input string, shortcut string"


class RegisterExecuted(BaseModel):
    key: str
    content: str



from python_search.config import ConfigurationLoader
from python_search.ranking.next_item_predictor.inference.inference import \
    Inference
from python_search.ranking.next_item_predictor.inference.input import \
    InferenceInput


class OfflineEvaluation:
    """
    Evaluate the model with a part of the training data
    """

    def run(self, model, dataset, X_test):
        """
        Computes the average position of the entray in the validation set
        """
        self._configuration = ConfigurationLoader().load_config()
        print("Starting offline evaluation")
        ids = [int(x) for x in X_test[:, 0].tolist()]
        df = dataset.toPandas()
        test_df = df[df["entry_number"].isin(ids)]
        print("TestDF shape: ", test_df.shape)

        inference = Inference(model=model)

        total_found = 0
        number_of_tests = 20
        avg_position = 0
        number_of_existing_keys = len(self._configuration.commands.keys())
        for index, row in test_df.iterrows():

            if (
                not self._key_exists(row["key"])
                or not self._key_exists(row["previous_key"])
                or not self._key_exists(row["previous_previous_key"])
            ):
                print(
                    f"Key pair does not exist any longer ({row['previous_key']}, {row['key']})"
                )
                continue

            input = InferenceInput(
                hour=row["hour"],
                month=row["month"],
                previous_key=row["previous_key"],
                previous_previous_key=row["previous_previous_key"],
            )
            result = inference.get_ranking(predefined_input=input, print_weights=False)

            metadata = {
                "pair": row["previous_key"] + " -> " + row["key"],
                "position_target": result.index(row["key"]),
                "Len": len(result),
                "type of result": type(result),
            }
            print(metadata)

            avg_position += metadata["position_target"]
            total_found += 1
            if total_found == number_of_tests:
                break

        avg_position = avg_position / number_of_tests
        result = {
            "avg_position_for_tests": avg_position,
            "number_of_tests": number_of_tests,
            "number_of_existing_keys": number_of_existing_keys,
        }
        print(result)

        return result

    def _key_exists(self, key):
        return key in self._configuration.commands.keys()



import json
import logging
from datetime import datetime

from colorama import Fore, Style

from python_search.config import ConfigurationLoader


class Preview:
    """
    Preview the entry in the search ui
    """

    def __init__(self):
        self.configuration = ConfigurationLoader()
        self.logger = logging.getLogger("preview_entry")
        # do not send the errors to stderr, in the future we should send to kibana or a file
        self.logger.propagate = False

    def display(self, entry_text: str):
        """
        Prints the entry in the preview window
        """
        data = entry_text.split(":")
        key = data[0]
        # the entry content is after the key + a ":" character
        serialized_content = entry_text[len(key) + 1 :]
        type = "Unknown"

        try:

            entry_data = self.configuration.load_entries()[key]
            print("")
            if "url" in entry_data or "file" in entry_data:

                value = entry_data.get("url", entry_data.get("file"))

                print(f"{Fore.GREEN}{value}{Style.RESET_ALL}")
                if "url" in entry_data:
                    del entry_data["url"]

                if "file" in entry_data:
                    del entry_data["file"]
                type = "UrlInterpreter" if "url" in entry_data else "File"

            if "snippet" in entry_data:
                print(f"{Fore.BLUE}{entry_data['snippet']}{Style.RESET_ALL}")
                del entry_data["snippet"]
                type = "Snippet"

            if "cli_cmd" in entry_data or "cmd" in entry_data:
                value = entry_data.get("cli_cmd", entry_data.get("cmd"))
                print(f"{Fore.RED}{value}{Style.RESET_ALL}")
                type = "Cmd" if "cmd" in entry_data else "CliCmd"

                if "cli_md" in entry_data:
                    del entry_data["cli_cmd"]

                if "cmd" in entry_data:
                    del entry_data["cmd"]

            if "callable" in entry_data:
                value = entry_data.get("callable")
                import dill

                print(f"{Fore.RED}{dill.source.getsource(value)}{Style.RESET_ALL}")
                type = "Callable"
                if "callable" in entry_data:
                    del entry_data["callable"]

            print("")

            print(f"key: {Fore.YELLOW}{key}{Style.RESET_ALL}")
            print("type: " + type)

            for key, value in entry_data.items():
                print(f"{key}: {value}")

            if "created_at" in entry_data:
                from dateutil import parser

                creation_date = parser.parse(entry_data["created_at"])
                today = datetime.now()
                print(f"entry age: {today - creation_date}")

            try:
                decoded_content = json.loads(serialized_content)
                print(f"position: {decoded_content['position']}")
            except Exception as e:
                self.logger.error(str(e))
                pass

        except BaseException as e:
            print(e)
            self.logger.error(str(e))



#!/usr/bin/env python3

from pyspark.sql import SparkSession
from pyspark.sql import functions as F

from python_search.config import DataConfig
from python_search.events.events import SearchRunPerformed

default_port = "9092"
host = f"localhost:{default_port}"


def consume_search_run_performed():
    """Entrypoint for the consumer"""
    SparkEventConsumer().from_class(SearchRunPerformed)


class SparkEventConsumer:
    """
    Listen to kafka events and store then in the data-wharehouse
    """

    def __init__(self, disable_await_termination=False):
        # by default awaits termination
        self.await_termination = not disable_await_termination

    def from_class(self, class_reference):
        topic_name = class_reference.__name__
        self.consume(topic_name, class_reference.get_schema())

    def consume(self, topic_name, schema):
        """
        To trigger run the following:

        spark-submit \
        --packages org.apache.spark:spark-sql-kafka-0-10_2.12:3.0.1 \
        $HOME/projects/PythonSearch/python_search/events/consumer.py consume_search_run_performed
        """

        spark = SparkSession.builder.getOrCreate()
        df = (
            spark.readStream.format("kafka")
            .option("kafka.bootstrap.servers", host)
            .option("failOnDataLoss", "false")
            .option("subscribe", topic_name)
            .option("subscribe", topic_name)
            .load()
        )
        # process the data here
        df = df.selectExpr("CAST(value as STRING) as value_decoded", "timestamp")
        df = df.withColumn(
            "value_as_json",
            F.from_json("value_decoded", schema),
        )
        df = df.select("value_as_json.*", "timestamp")

        streamingQuery = (
            df.writeStream.format("parquet")
            .outputMode("append")
            .option(
                "path", f"{DataConfig.DATA_WAREHOUSE_FOLDER}/dataframes/{topic_name}"
            )
            .option(
                "checkpointLocation",
                f"{DataConfig.DATA_WAREHOUSE_FOLDER}/checkpoints/{topic_name}",
            )  # write-ahead logs for
            .trigger(processingTime="1 second")
            .start()
        )

        if self.await_termination:
            streamingQuery.awaitTermination()


if __name__ == "__main__":
    import fire

    fire.Fire()



from __future__ import annotations

from python_search.environment import is_mac


class SupportedBrowsers:
    CHROME = "chrome"
    FIREFOX = "firefox"


class Browser:
    """
    Abstracts the browser features cross-platform
    """

    # a tuple with the binary and the type
    _MAC_DEFAULT_BROWSER = (
        '/usr/bin/open -a "/Applications/Google Chrome.app"',
        SupportedBrowsers.CHROME,
    )
    _LINUX_DEFAULT_BROWSER = "google-chrome", SupportedBrowsers.CHROME

    def open(self, url: str, app_mode=False, incognito=False) -> None:
        """
        performs the open
        """
        params = locals()
        del params["self"]
        cmd_to_run = self.open_cmd(**params)
        print("Comand to run:", cmd_to_run)
        from python_search.interpreter.cmd import CmdInterpreter

        CmdInterpreter({"cmd": cmd_to_run}).interpret_default()

    def open_cmd(self, url: str, app_mode=False, incognito=False) -> str:
        """
        Returns the shell command to open the browser
        """

        if is_mac():
            browser, type = self._MAC_DEFAULT_BROWSER
        else:
            browser, type = self._LINUX_DEFAULT_BROWSER

        # @todo this not always work consider adding a a retry mechanism
        # if type == SupportedBrowsers.CHROME and app_mode:
        #    return f"/Applications/Google\ Chrome.app/Contents/MacOS/Google\ Chrome  --app='{url}'"

        cmd = f"{browser} '{url}'"

        if type == SupportedBrowsers.CHROME and incognito:
            cmd = f"/Applications/Google\ Chrome.app/Contents/MacOS/Google\ Chrome  --incognito '{url}'"

        return cmd


def main():
    import fire

    fire.Fire(Browser)


if __name__ == "__main__":
    main()



from grimoire.shell import shell

from python_search.apps.clipboard import Clipboard
from python_search.exceptions import CommandDoNotMatchException
from python_search.interpreter.base import BaseInterpreter


class SnippetInterpreter(BaseInterpreter):
    """
    Snippet handler
    @todo rename to entry.Snippet
    """

    def __init__(self, cmd, context=None):
        self.context = context

        if type(cmd) == str:
            self.cmd = {"snippet": cmd}
            return

        if type(cmd) is dict and "snippet" in cmd:
            self.cmd = cmd
            return

        raise CommandDoNotMatchException(
            f"Not Valid {self.__class__.__name__} command {cmd}"
        )

    def interpret_default(self):
        Clipboard().set_content(self.cmd["snippet"])

        if "type-it" in self.cmd:
            snippet = self.apply_directory(self.cmd["snippet"])
            shell.run(f"setxkbmap ; xdotool type '{snippet}'")
            shell.run(f"xdotool key Return ")

        return

    def copiable_part(self):
        return self.cmd["snippet"]



import os
from typing import Any, Tuple

import mlflow
import numpy as np
from keras import layers
from keras.models import Sequential
from sklearn.model_selection import train_test_split

from python_search.config import DataConfig
from python_search.ranking.next_item_predictor.offline_evaluation import \
    OfflineEvaluation
from python_search.ranking.next_item_predictor.training_dataset import \
    TrainingDataset
from python_search.ranking.next_item_predictor.transform import Transform


class Train:
    EPOCHS = 30
    TEST_SPLIT_SIZE = 0.10
    BATCH_SIZE = 128

    def __init__(self, epochs=None):
        if not epochs:
            epochs = Train.EPOCHS
        self.epochs = epochs
        # enable the profiling scaffolding
        os.environ["TIME_IT"] = "1"

    def train_and_log(self, dataset):
        """
        train the model and log it to MLFlow
        """

        mlflow.set_tracking_uri(f"file:{DataConfig.MLFLOW_MODELS_PATH}")
        # this creates a new experiment
        mlflow.set_experiment(DataConfig.NEXT_ITEM_EXPERIMENT_NAME)
        mlflow.keras.autolog()
        # @todo: try mlflow.keras.autolog()

        with mlflow.start_run():
            model, metrics, offline_evaluation = self.train(dataset)
            mlflow.log_params(metrics)
            mlflow.log_params(offline_evaluation)

        return model, metrics, offline_evaluation

    def train(self, dataset: TrainingDataset, plot_history=False):
        """
        performs training

        :param dataset:
        :param plot_history:
        :return:
        """

        # prepare the data
        X, Y = Transform().transform_train(dataset)
        X_train, X_test, Y_train, Y_test = self.split(X, Y)

        X_test = np.where(np.isnan(X_test), 0.5, X_test)
        Y_test = np.where(np.isnan(Y_test), 0.5, Y_test)
        Y_train = np.where(np.isnan(Y_train), 0.5, Y_train)

        X_test_p = np.delete(X_test, 0, axis=1)
        X_train_p = np.delete(X_train, 0, axis=1)

        model, history = self._only_train(X_train_p, X_test_p, Y_train, Y_test)
        metrics = self._compute_standard_metrics(
            model, X_train_p, X_test_p, Y_train, Y_test
        )

        offline_evaluation = OfflineEvaluation().run(model, dataset, X_test)

        if plot_history:
            self._plot_training_history(history)

        return model, metrics, offline_evaluation

    def _only_train(self, X_train, X_test, Y_train, Y_test) -> Tuple[Any, Any]:

        print("Starting train with N epochs, N=", self.epochs)
        print(
            {
                "shape_x_train": X_train.shape,
                "shape_x_test": X_test.shape,
                "shape_y_train": Y_train.shape,
                "shape_y_test": Y_test.shape,
                "X_train has nan: ": np.any(np.isnan(X_train)),
                "Y_train has nan: ": np.any(np.isnan(Y_train)),
                "X_test has nan: ": np.any(np.isnan(X_test)),
                "y_test has nan: ": np.any(np.isnan(Y_test)),
            }
        )

        model = Sequential()
        model.add(layers.Dense(128, activation="relu"))
        model.add(layers.Dropout(0.5))
        model.add(layers.Dense(64, activation="relu"))
        model.add(layers.Dropout(0.5))
        model.add(layers.Dense(1))
        model.compile(optimizer="rmsprop", loss="mse", metrics=["mae", "mse"])

        history = model.fit(
            X_train,
            Y_train,
            epochs=self.epochs,
            batch_size=Train.BATCH_SIZE,
            validation_data=(X_test, Y_test),
        )

        return model, history

    def split(self, X, Y):
        X_train, X_test, Y_train, Y_test = train_test_split(
            X, Y, test_size=Train.TEST_SPLIT_SIZE, random_state=42
        )
        return X_train, X_test, Y_train, Y_test

    def _compute_standard_metrics(self, model, X_train, X_test, Y_train, Y_test):
        """
        computes mse and mae for train and test splits
        """
        train_mse, train_mae, train_mse2 = model.evaluate(X_train, Y_train)
        test_mse, test_mae, test_mse2 = model.evaluate(X_test, Y_test)

        metrics = {
            "train_mse": train_mse,
            "train_mae": train_mae,
            "test_mse": test_mse,
            "test_mae": test_mae,
        }
        print("Metrics:", metrics)

        return metrics

    def _plot_training_history(self, history):
        import matplotlib.pyplot as plt

        loss = history.history["loss"]
        val_loss = history.history["val_loss"]

        epochs_range = range(1, self.epochs + 1)
        plt.plot(epochs_range, loss, "bo", label="Training Loss")
        plt.plot(epochs_range, val_loss, "b", label="Validation Loss")
        plt.title("Training and validation loss")
        plt.xlabel("Epochs")
        plt.ylabel("Loss")
        plt.legend()
        plt.show()

    def _normalize(self, X_train, X_test):
        # normalize
        mean = X_train.mean(axis=0)
        X_train -= mean
        std = X_train.std(axis=0)
        X_train /= std

        X_test -= mean
        X_test /= std

        return X_train, X_test



from python_search.features import FeatureToggle


class Report:
    """
    Reponsisble to give an overview of the current state of the application.
    """

    def generate(self):
        return {
            "features": {
                "ranking_latest_used": FeatureToggle().is_enabled("ranking_latest_used")
            }
        }



#!/usr/bin/env python
# using ipython interferes with fire arguments passing

import os
from typing import Optional

from pyspark.sql import SparkSession
from sklearn.metrics import mean_absolute_error

from python_search.infrastructure.performance import timeit
from python_search.observability.logger import initialize_logging
from python_search.ranking.next_item_predictor.evaluator import Evaluate
from python_search.ranking.next_item_predictor.train_keras import Train
from python_search.ranking.next_item_predictor.train_xgboost import \
    TrainXGBoost
from python_search.ranking.next_item_predictor.training_dataset import \
    TrainingDataset

initialize_logging()


class Pipeline:
    """
    Exposes the whole ML pipeline, the runs everything
    """

    def __init__(self):
        os.environ["TIME_IT"] = "1"

    def train(self, use_cache=True, log_model=True):
        """
        Trains both xgboost and keras models

        Args:
            use_cache:
            log_model:

        Returns:

        """
        parameters = locals()
        del parameters["self"]

        self.train_xgboost(**parameters)
        self.train_keras(**parameters)

    @timeit
    def train_keras(
        self,
        *,
        dataset: Optional[TrainingDataset] = None,
        epochs=None,
        use_cache=True,
        log_model=True,
    ):
        print("Start training Keras model")
        if not dataset:
            print(f"Using data with cache: {use_cache} type: {type(use_cache)}")
            dataset = self.build_dataset(use_cache=use_cache)

        print(f"Custom epochs {epochs}")

        if log_model:
            model, metrics, offline_evaluation = Train(epochs).train_and_log(dataset)
        else:
            model, metrics, offline_evaluation = Train(epochs).train(
                dataset, plot_history=True
            )

        print(
            {
                "metrics": metrics,
                "offline_evaluation": offline_evaluation,
            }
        )

    def train_xgboost(self, use_cache=True, log_model=True):
        """
        Train the XGBoost model
        """
        print("Start training XGBoost model")
        dataset = self.build_dataset(use_cache=use_cache)
        if log_model:
            TrainXGBoost().train_and_log(dataset)
        else:
            TrainXGBoost().train(dataset)

    @timeit
    def build_dataset(self, use_cache=False, view_only=False):
        """
        Builds the dataset ready for training
        """
        dataset = TrainingDataset().build(use_cache)

        print("Training dataset is ready, printing top 10", dataset.show(n=10))

        if view_only:
            print("View only enabled will early return")
            return

        return dataset

    def evaluate(self):
        """
        Evaluate the latest dataset
        """
        return Evaluate().evaluate

    def baseline_mse(self, dataset=None):
        """
        naive approach of setting the same as input and output, used as baseline to measure the real model against
        """
        import pyspark.sql.functions as F
        from sklearn.metrics import mean_squared_error

        if not dataset:
            dataset = self.build_dataset(use_cache=True)

        # apply only to the  ones with the same name in input and output
        # complete dataset 8k, with the same name in input and output 150
        naive = (
            dataset.filter("key == previous_key")
            .select("key", "label")
            .withColumn("baseline", F.lit(1))
        )
        pd_naive = naive.select("label", "baseline").toPandas()
        return {
            "baseline_mse": mean_squared_error(pd_naive.label, pd_naive.baseline),
            "baseline_mae": mean_absolute_error(pd_naive.label, pd_naive.baseline),
        }

    def _spark(self):
        return SparkSession.builder.getOrCreate()


def main():
    import fire

    fire.Fire(Pipeline)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
from typing import List, Tuple

import msgpack_numpy as m
import numpy as np
from numpy import ndarray

from python_search.config import ConfigurationLoader
from python_search.infrastructure.performance import timeit
from python_search.infrastructure.redis import PythonSearchRedis
from python_search.ranking.entries_loader import EntriesLoader


class RedisEmbeddingsReader:
    """
    Responsible for quickly reading the embeddings from redis
    """

    def __init__(self):
        self.client = PythonSearchRedis.get_client()

    @timeit
    def load(self, keys) -> dict[str, bytes]:
        """
        Returns a dictionary with the key being the embedding key and the value being the bytes of the value
        """
        pipe = self.client.pipeline()

        for key in keys:
            pipe.hget(f"k_{key}", "embedding")

        all_embeddings = pipe.execute()

        if len(all_embeddings) != len(keys):
            raise Exception(
                "Number of keys returned from redis does not match the number of embeddings found"
            )

        embedding_mapping = dict(zip(keys, all_embeddings))
        return embedding_mapping

    def load_all_keys(self):
        """
        Generate embeddings for all currently existing entries
        """
        keys = EntriesLoader.load_all_keys()

        return self.load(keys)


class RedisEmbeddingsWriter:
    """Responsible for writing the embeddings in redis"""

    def __init__(self):
        self.client = PythonSearchRedis.get_client()

    def sync_missing(self):
        """ "
        Write embeddings of keys not present in redis
        """

        # load current embeddings

        existing_embeddings = RedisEmbeddingsReader().load_all_keys()

        empty_keys = [
            key for key, embedding in existing_embeddings.items() if embedding is None
        ]

        if not len(empty_keys):
            print("No keys missing sync, exit early")
            return

        embeddings = create_key_indexed_embedding(empty_keys)

        for key, embedding in embeddings.items():
            self.write_embedding(key, embedding)

        print("Done!")

    def sync_all(self):
        """
        Generate embeddings for all currently existing entries
        """
        keys = EntriesLoader.load_all_keys()
        embeddings = create_key_indexed_embedding(keys)

        for key, embedding in embeddings.items():
            self.write_embedding(key, embedding)

        print("Done!")

    def write_embedding(self, key: str, embedding: np.ndarray):
        self.client.hset(
            f"k_{key}", "embedding", EmbeddingSerialization.serialize(embedding)
        )

    def read_embedding(self, key):
        return EmbeddingSerialization.read(self.client.hget(key, "embedding"))


class EmbeddingSerialization:
    """Responsible to encode the numpy embeddings in a format readis can read and write from and to"""

    @staticmethod
    def read(embedding):
        return m.unpackb(embedding)

    @staticmethod
    def serialize(embedding):
        return m.packb(embedding)


def create_embeddings_from_strings(keys: List[str]) -> ndarray:
    from sentence_transformers import SentenceTransformer

    transformer = SentenceTransformer("nreimers/MiniLM-L6-H384-uncased")
    return transformer.encode(keys, batch_size=128, show_progress_bar=True)


def create_key_indexed_embedding(keys) -> dict[str, str]:
    """
    Create an embedding dict
    """
    unique_keys = list(set(keys))

    entries = ConfigurationLoader().load_entries()
    unique_bodies = []

    for key in unique_keys:
        if key in entries:
            body = str(entries[key])
            print(f"For key '{key}', found body to encode: {body}")
            unique_bodies.append(key + " " + body)
        else:
            print(f"Could not find body for key: {key}")
            unique_bodies.append(key)

    embeddings = create_embeddings_from_strings(unique_bodies)
    embeddings_keys = dict(zip(unique_keys, embeddings))
    return embeddings_keys


def main():
    import fire

    fire.Fire()


if __name__ == "__main__":
    main()



from __future__ import annotations

import json
import logging
from typing import Optional

from python_search.apps.clipboard import Clipboard
from python_search.apps.window_manager import I3
from python_search.context import Context
from python_search.environment import is_mac


class BaseInterpreter:
    """parent of all _interpreters, Cannot instantiate directly"""

    def __init__(self, cmd, context: Optional[Context] = None):
        self.cmd = cmd
        self.context = context

    def default(self) -> None:
        if "ask_confirmation" in self.cmd and not self._confirmed_continue():
            return

        if "call_after" in self.cmd or "call_before" in self.cmd:
            logging.info("Enabled sequential execution flag enabled")
            self.context.enable_sequential_execution()

        if "disable_sequential_execution" in self.cmd:
            self.context.disable_sequential_execution()
            logging.info("Disable sequential execution flag enabled")

        self._call_before()
        result = self.interpret_default()
        self._call_after()

        return result

    def _confirmed_continue(self) -> bool:
        from python_search.entry_capture.ui import AskQuestion

        result = AskQuestion().ask(
            f"Type (y) if you wanna to proceed to run command: {self.cmd['cmd']}"
        )

        if result == "y":
            return True

        from python_search.apps.notification_ui import send_notification

        send_notification(f"Operation cancelled. Confirmation response was '{result}'")

        return False

    def _call_before(self):
        if "call_before" not in self.cmd:
            return

        logging.info("Call before enabled")
        logging.info(f"Executing post-processing cmd {self.cmd['call_before']}")
        self.context.get_interpreter().default(self.cmd["call_before"])

    def _call_after(self):
        if "call_after" not in self.cmd:
            return

        logging.info("Call after enabled")
        logging.info(f"Executing post-processing cmd {self.cmd['call_after']}")
        self.context.get_interpreter().default(self.cmd["call_after"])

    def interpret_default(self):
        raise Exception("Implement me!")

    def interpret_clipboard(self):
        return Clipboard().set_content(self.copiable_part())

    def copiable_part(self):
        return self.serialize()

    def serialize(self):
        return str(self.cmd)

    def to_dict(self):
        return self.cmd

    def serialize_entry(self):
        return json.dumps(self.cmd)

    def apply_directory(self, cmd):
        if "directory" in self.cmd:
            cmd = f'cd {self.cmd["directory"]} ; {cmd}'
        return cmd

    def try_to_focus(self) -> bool:
        """
        Focus on the window if it is already running and return True.
        If it is not running, return False.
        """

        if "focus_match" not in self.cmd:
            return False

        if is_mac():
            print("Will not try to focus as this is not supported yet")
            return False

        return I3().focus_on_window_with_title(self.cmd["focus_match"])



from __future__ import annotations

import os
import traceback
from typing import Any, List, Optional

from python_search.config import ConfigurationLoader, PythonSearchConfiguration
from python_search.infrastructure.performance import timeit
from python_search.ranking.models import PythonSearchMLFlow
from python_search.ranking.next_item_predictor.inference.input import \
    InferenceInput
from python_search.ranking.next_item_predictor.transform import Transform


class Inference:
    """
    Performs the ranking inference on all existing keys in the moment
    """

    PRODUCTION_RUN_ID = "78dacecfcab544e391e76ff25a21c012"

    def __init__(
        self,
        configuration: Optional[PythonSearchConfiguration] = None,
        run_id: Optional[str] = None,
        model: Optional[Any] = None,
    ):

        self.debug = os.getenv("DEBUG", False)
        self.run_id = run_id if run_id else self.PRODUCTION_RUN_ID
        if "FORCE_RUN_ID" in os.environ:
            self.run_id = os.environ["FORCE_RUN_ID"]

        if model:
            print("Using custom passed model")
        else:
            print("Using run id: " + self.run_id)
        configuration = (
            configuration if configuration else ConfigurationLoader().load_config()
        )
        # previous key should be setted in runtime
        self.previous_key = None
        self.all_keys = configuration.commands.keys()
        self._transform = Transform()

        self.model = model if model else self._load_mlflow_model(run_id=self.run_id)

    @timeit
    def get_ranking(
        self,
        predefined_input: Optional[InferenceInput] = None,
        print_weights=False,
        return_weights=False,
    ) -> List[str]:
        """
        Gets the ranking from the next item model
        """
        print("Number of existing keys: ", str(len(self.all_keys)))
        inference_input = (
            predefined_input
            if predefined_input
            else InferenceInput.with_keys(
                self._transform.inference_embeddings.get_recent_key_with_embedding(),
                self._transform.inference_embeddings.get_recent_key_with_embedding(
                    second_recent=True
                ),
            )
        )

        try:
            X = self._transform.transform_inference(inference_input, self.all_keys)
            Y = self._predict(X)
            print(Y)
            result = list(zip(self.all_keys, Y))
            result.sort(key=lambda x: x[1], reverse=True)
            if return_weights:
                return result
            if print_weights:
                print(result)

            only_keys = [entry[0] for entry in result]
            print("Ranking inference succeeded")
        except Exception as e:
            print(
                "Error while performing inference, returning baseline ranking. Details: "
                + e.__str__()
            )

            print(traceback.format_exc())
            only_keys = self.all_keys

        print("ONlye keys: ", only_keys)
        return only_keys

    @timeit
    def _predict(self, X):
        return self.model.predict(X)

    @timeit
    def _load_mlflow_model(self, run_id=None):
        model = PythonSearchMLFlow().get_next_predictor_model(run_id=run_id)
        return model



from fastapi import FastAPI
from fastapi.responses import PlainTextResponse

app = FastAPI()
from python_search.config import ConfigurationLoader
from python_search.ranking.ranking import RankingGenerator

config = ConfigurationLoader().load_config()
generator = RankingGenerator(config)


@app.get("/ranking/generate", response_class=PlainTextResponse)
def generate_ranking():
    global generator
    return generator.generate()


@app.get("/ranking/reload_and_generate", response_class=PlainTextResponse)
def reload():
    global generator
    global config
    del config
    del generator
    config = ConfigurationLoader().reload()
    generator = RankingGenerator(config)
    return generator.generate()


@app.get("/_health")
def health():
    global generator

    from python_search.events.latest_used_entries import LatestUsedEntries

    entries = LatestUsedEntries().get_latest_used_keys()

    return {
        "keys_count": len(ConfigurationLoader().load_config().commands.keys()),
        "run_id": generator.inference.PRODUCTION_RUN_ID,
        "latest_used_entries": entries,
    }


def main():
    import os

    import uvicorn

    os.putenv("WEB_CONCURRENCY", "1")
    reload = False
    if "PS_DEBUG" in os.environ:
        print("Debug mode is ON, enabling reload")
        reload = True
    else:
        print("Debug mode is OFF")

    uvicorn.run("python_search.web_api:app", host="0.0.0.0", port=8000, reload=reload)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['python_search',
 'python_search.apps',
 'python_search.data_ui',
 'python_search.datasets',
 'python_search.entry_capture',
 'python_search.events',
 'python_search.infrastructure',
 'python_search.init_project',
 'python_search.interpreter',
 'python_search.observability',
 'python_search.ranking',
 'python_search.ranking.next_item_predictor',
 'python_search.ranking.next_item_predictor.inference',
 'python_search.search_ui',
 'python_search.shortcut']

package_data = \
{'': ['*']}

install_requires = \
['PySimpleGUI>=4.60.1,<5.0.0',
 'PyYAML>=6.0,<7.0',
 'certifi>=2022.6.15,<2023.0.0',
 'colorama>=0.4.5,<0.5.0',
 'fastapi>=0.79.0,<0.80.0',
 'fire>=0.4.0,<0.5.0',
 'kafka-python>=2.0.2,<3.0.0',
 'matplotlib>=3.5.2,<4.0.0',
 'msgpack-numpy>=0.4.8,<0.5.0',
 'numpy>=1.23.0,<1.24.0',
 'pandas>=1.4.3,<2.0.0',
 'personal-grimoire==1.6',
 'pydantic>=1.9.1,<2.0.0',
 'redis>=4.3.4,<5.0.0',
 'scipy>=1.8.1',
 'uvicorn>=0.18.2,<0.19.0',
 'xgboost>=1.6.1,<2.0.0']

extras_require = \
{'all': ['pyspark>=3.3.0,<4.0.0']}

entry_points = \
{'console_scripts': ['browser = python_search.apps.browser:main',
                     'clipboard = python_search.apps.clipboard:main',
                     'collect_input = python_search.apps.collect_input:main',
                     'entry_embeddings = '
                     'python_search.ranking.entry_embeddings:main',
                     'next_item_pipeline = '
                     'python_search.ranking.next_item_predictor.pipeline:main',
                     'notify_send = python_search.apps.notification_ui:main',
                     'python_search = python_search.cli:main',
                     'python_search_infra = '
                     'python_search.infrastructure.infrastructure:main',
                     'python_search_webapi = python_search.web_api:main',
                     'run_entry = python_search.cli:_run_key_bin']}

setup_kwargs = {
    'name': 'python-search',
    'version': '0.5.7',
    'description': 'Build your knowledge database in python and retrieve it efficiently',
    'long_description': 'None',
    'author': 'Jean Carlo Machado',
    'author_email': 'machado.c.jean@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



from grimoire.shell import shell
from grimoire.string import chomp, remove_new_lines

from python_search.environment import is_mac


class Clipboard:
    def get_content(self, source="--primary"):
        """
        Accepted values are --primary and --clipboard
        """

        cmd = f"xsel {source} --output"
        if is_mac():
            cmd = "pbpaste"

        result = shell.run_with_result(cmd)
        result = chomp(result)

        return result

    def get_content_preview(self):
        content = self.get_content()
        content = content.strip(" \t\n\r")
        content = remove_new_lines(content)
        content_len = len(content)
        desized_preview_size = 10
        size_of_preview = (
            desized_preview_size if desized_preview_size < content_len else content_len
        )

        final_content = content[0:size_of_preview]
        suffix = " ..." if len(content) > size_of_preview else ""
        return f"{final_content}{suffix}"

    def set_content(self, content: str, enable_notifications=True):
        """
        Put a string in the clibboard
        :param content:
        :param enable_notifications:
        :return:
        """

        def shellquote(s):
            return "'" + s.replace("'", "'\\''") + "'"

        sanitized = shellquote(content)

        clipboard_cmd = "xsel --clipboard --primary --input"
        if is_mac():
            clipboard_cmd = "pbcopy"

        cmd = f"echo {sanitized} | {clipboard_cmd}"

        if enable_notifications:
            from python_search.apps.notification_ui import send_notification

            send_notification(f"Content copied: {sanitized}")

        return shell.run(cmd)


def main():
    import fire

    fire.Fire(Clipboard)


if __name__ == "__main__":
    main()



from __future__ import annotations

import datetime
import os

from python_search.ranking.next_item_predictor.inference.embeddings_loader import \
    InferenceEmbeddingsLoader


class InferenceInput:
    hour: int
    month: int
    previous_key: str
    previous_previous_key: str

    def __init__(self, *, hour, month, previous_key, previous_previous_key):
        self.hour = hour
        self.month = month
        self.previous_key = previous_key
        self.previous_previous_key = previous_previous_key

    @staticmethod
    def with_keys(previous_key: str, previous_previous_key: str) -> "InferenceInput":
        """
        Do inference based on the current time and the recent used keys
        """
        now = datetime.datetime.now()
        hour = int(os.environ["FORCE_HOUR"]) if "FORCE_HOUR" in os.environ else now.hour
        month = (
            int(os.environ["FORCE_MONTH"]) if "FORCE_MONTH" in os.environ else now.month
        )

        instance = InferenceInput(
            hour=hour,
            month=month,
            previous_key=previous_key,
            previous_previous_key=previous_previous_key,
        )

        print("Inference input: ", instance.__dict__)

        return instance
